import warnings
from typing import Any


class Response:
    """Class to format the response, so that it can be accessed as an object variable.

    >>> Response

    """

    def __init__(self, dictionary: dict):
        """Extracts the properties ``ok``, ``status`` and ``body`` from a dictionary.

        Args:
            dictionary: Takes a dictionary of key-value pairs as an argument.
        """
        self.raw: dict = dictionary
        self._ok: bool = dictionary.get('ok')
        self._status: int = dictionary.get('status')
        self._body: Any = dictionary.get('body')
        self._count: int or None = dictionary.get('count')

    @property
    def ok(self) -> bool:
        """Returns the extracted class variable.

        Returns:
            bool:
            ``True`` or ``False`` based on the arg value received.
        """
        return self._ok

    @property
    def status(self) -> int:
        """Returns the extracted class variable.

        Returns:
            int:
            ``HTTP`` status code as received.
        """
        return self._status

    @property
    def body(self) -> str:
        """Returns the extracted class variable.

        Returns:
            str:
            Returns the message as received.
        """
        return self._body

    def json(self) -> dict:
        """Returns a dictionary of the argument that was received during class initialization.

        Returns:
            dict:
            Returns the dictionary received as arg.
        """
        return self.raw

    @property
    def count(self) -> int:
        """Takes the number of un-read emails and returns it in a class. Only works for read email.

        Returns:
            int:
            Returns the number of unread emails.
        """
        if not isinstance(self._count, int):
            warnings.warn('count property is valid only for the class ReadEmail')
        return self._count



import subprocess
from typing import List

from .exceptions import InvalidDomain, NotMailServer
from .support import hostname_to_ip, matrix_to_flat_list, remove_duplicates


def get_mx_records(domain: str) -> List:
    """Get MX (Mail Exchange server) records for the given domain.

    Args:
        domain: FQDN (Fully Qualified Domain Name) extracted from the email address.

    Returns:
        list:
        List of IP addresses of all the mail exchange servers from authoritative/non-authoritative answer section.
    """
    try:
        output = subprocess.check_output(f'nslookup -q=mx {domain}', shell=True)
        result = [hostname_to_ip(hostname=line.split()[-1]) for line in output.decode().splitlines()
                  if line.startswith(domain)]
        cleaned = remove_duplicates(input_=matrix_to_flat_list(input_=result))
        if not cleaned:
            raise NotMailServer(f"Domain {domain!r} is not a mail server.")
        return cleaned
    except subprocess.CalledProcessError:
        raise InvalidDomain(f"Domain {domain!r} is invalid.")



import os
from smtplib import SMTP, SMTPAuthenticationError, SMTPConnectError
from threading import Thread

from dotenv import load_dotenv

from .responder import Response
from .sms_deleter import DeleteSent

if os.path.isfile('.env'):
    load_dotenv(dotenv_path='.env')


class Messenger:
    """Initiates Messenger object to send an SMS to a phone number using SMS gateway provided by the mobile carrier.

    >>> Messenger

    """

    SMS_GATEWAY = {
        "at&t": "mms.att.net",
        "t-mobile": "tmomail.net",
        "verizon": "vtext.com",
        "boost": "smsmyboostmobile.com",
        "cricket": "sms.cricketwireless.net",
        "us-cellular": "email.uscc.net",
    }

    def __init__(self, gmail_user: str = os.environ.get('gmail_user') or os.environ.get('GMAIL_USER'),
                 gmail_pass: str = os.environ.get('gmail_pass') or os.environ.get('GMAIL_PASS')):
        """Initiates all the necessary args.

        Args:
            gmail_user: Gmail username to authenticate SMTP lib.
            gmail_pass: Gmail password to authenticate SMTP lib.

        See Also:
            Carrier defaults to ``t-mobile`` which uses ``tmomail.net`` as the SMS gateway.
        """
        self.server, self.mail = None, None
        if not all([gmail_user, gmail_pass]):
            raise ValueError(
                'Cannot proceed without the args: `gmail_user` and `gmail_pass`'
            )
        if not gmail_user.endswith('@gmail.com'):
            gmail_user = gmail_user + '@gmail.com'
        self.gmail_user = gmail_user
        self.gmail_pass = gmail_pass
        self.server = SMTP(host="smtp.gmail.com", port=587)
        self._authenticated = False

    @property
    def authenticate(self) -> Response:
        """Starts the TLS server and authenticates the user.

        Returns:
            Response:
            A custom response class with properties: ok, status and body to the user.
        """
        self.server.starttls()
        try:
            self.server.login(user=self.gmail_user, password=self.gmail_pass)
            self._authenticated = True
            return Response(dictionary={
                'ok': True,
                'status': 200,
                'body': "Authentication was successful."
            })
        except SMTPAuthenticationError:
            return Response(dictionary={
                'ok': False,
                'status': 401,
                'body': "GMAIL login failed with SMTPAuthenticationError: Username and Password not accepted.\n"
                        "Ensure the credentials stored in env vars are set correct.\n"
            })
        except SMTPConnectError:
            return Response(dictionary={
                'ok': False,
                'status': 503,
                'body': "Error during connection establishment with GMAIL server."
            })

    def __del__(self):
        """Destructor has been called to close the TLS connection and logout."""
        if self.server:
            self.server.close()

    @staticmethod
    def _validate_phone(phone):
        """Validates all the arguments passed during object initialization.

        Raises:
            ValueError: If any arg is missing or malformed.
        """
        if len(phone) != 10 and len(phone) != 12:
            raise ValueError('Phone number should either be 10 or 12 digits (if includes country code)')
        if phone.startswith('+') and not phone.startswith('+1'):
            raise ValueError('Unsupported country code. Module works only for US based contact numbers.')

    @staticmethod
    def _generate_address(phone, gateway) -> str:
        """Validates the digits in the phone number and forms the endpoint using the phone number and sms gateway.

        Returns:
            str:
            Returns the formed endpoint. `Example: ``+11234567890@tmomail.net```
        """
        if len(phone) == 11 and phone.startswith('1'):
            phone = f'+{phone}'
        if not phone.startswith('+'):
            phone = f'+1{phone}'
        return phone + '@' + gateway

    def send_sms(self, message: str, phone: str = os.environ.get('phone') or os.environ.get('PHONE'),
                 subject: str = None, carrier: str = 't-mobile', sms_gateway: str = None,
                 delete_sent: bool = True) -> Response:
        """Initiates a TLS connection and sends a text message through SMS gateway of destination number.

        Args:
            phone: Phone number stored as env var.
            message: Content of the message.
            subject: Subject line for the message. Defaults to "Message from GmailConnector"
            carrier: Takes any of ``at&t``, ``t-mobile``, ``verizon``, ``boost``, ``cricket``, ``us-cellular``
            sms_gateway: Takes the SMS gateway of the carrier as an argument.
            delete_sent: Boolean flag to delete the message from GMAIL's sent items. Defaults to ``True``.

        Raises:
            UnicodeEncodeError

        See Also:
            - Encodes body of the message to `ascii` with `ignore` flag and then decodes it.
            - This is done to ignore special characters (like °) without raising `UnicodeEncodeError`
            - Validates the endpoint ``phone-number@sms-gateway`` before trying to send the SMS.

        Notes:
            Other flags that can be set includes `replace` and `xmlcharrefreplace`

        Returns:
            Response:
            A custom response class with properties: ok, status and body to the user.
        """
        carrier = carrier.lower()
        if carrier not in list(self.SMS_GATEWAY.keys()):
            carrier = 't-mobile'
        gateway = sms_gateway or self.SMS_GATEWAY.get(carrier)

        self._validate_phone(phone=phone)
        body = f'\n\n{message}'.encode('ascii', 'ignore').decode('ascii')
        subject = subject or f"Message from {self.gmail_user.replace('@gmail.com', '')}"
        to = self._generate_address(phone=phone, gateway=gateway)
        if not self._authenticated:
            status = self.authenticate
            if not status.ok:
                return status
        message = (f"From: {self.gmail_user}\n" + f"To: {to}\n" + f"Subject: {subject}\n" + body)
        if len(message) > 428:
            return Response(dictionary={
                'ok': False,
                'status': 413,
                'body': f'Payload length: {len(message):,}, which is more than the optimal size: 428. '
                        f'Message length: {len(body):,}'
            })

        self.server.sendmail(self.gmail_user, to, message)

        if delete_sent:
            delete = DeleteSent(username=self.gmail_user, password=self.gmail_pass, subject=subject)
            Thread(target=delete.delete_sent, daemon=True).start()

        return Response(dictionary={
            'ok': True,
            'status': 200,
            'body': f'SMS has been sent to {to}'
        })



from ipaddress import IPv4Address, IPv6Address
from typing import Union

import idna.core

from .exceptions import AddressFormatError


class ValidateAddress:
    """Initiates ValidateAddress object to split the address into user and domin to further validate.

    >>> ValidateAddress

    """

    def __init__(self, address: str):
        self._address = address
        try:
            self._user, self._domain = self._address.rsplit('@', 1)
        except ValueError:
            raise AddressFormatError
        if not self._user:
            raise AddressFormatError("Empty user")
        try:
            self._domain = idna.core.encode(self._domain).decode('ascii')
        except idna.core.IDNAError as error:
            raise AddressFormatError(error)

    @property
    def user(self) -> str:
        """Returns only the user part of the address."""
        return self._user

    @property
    def domain(self) -> Union[str, IPv4Address, IPv6Address]:
        """Returns only the domain part of the address."""
        return self._domain

    @property
    def email(self) -> str:
        """Returns the email address."""
        return '@'.join((self.user, self.domain))



class AddressFormatError(ValueError):
    """Custom error object for invalid address format."""


class NotMailServer(ValueError):
    """Custom error object for domain that is not a mail server."""


class InvalidDomain(ValueError):
    """Custom error object for invalid domain."""



import os
from datetime import datetime, timedelta, timezone
from email import message_from_bytes
from email.header import decode_header, make_header
from imaplib import IMAP4_SSL

import pytz
from dotenv import load_dotenv

from .responder import Response

if os.path.isfile('.env'):
    load_dotenv(dotenv_path='.env')


class ReadEmail:
    """Initiates Emailer object to authenticate and yield the unread emails.

    >>> ReadEmail

    """

    FOLDERS = ['inbox', '"[Gmail]/Important"', '"[Gmail]/Starred"', '"[Gmail]/All Mail"', '"[Gmail]/Sent Mail"',
               '"[Gmail]/Drafts"', '"[Gmail]/Spam"', '"[Gmail]/Trash"']
    CATEGORIES = ['ALL', 'SEEN', 'UNSEEN']
    ADDITIONAL_CATEGORIES = ['SMALLER', 'TEXT', 'SINCE']
    LOCAL_TIMEZONE = datetime.now(timezone.utc).astimezone().tzinfo

    def __init__(self, gmail_user: str = os.environ.get('gmail_user') or os.environ.get('GMAIL_USER'),
                 gmail_pass: str = os.environ.get('gmail_pass') or os.environ.get('GMAIL_PASS'),
                 folder: str = 'inbox'):
        """Initiates all the necessary args.

        Args:
            gmail_user: Login email address.
            gmail_pass: Login password.
            folder: Folder where the unread emails have to be read from.

        References:
            https://imapclient.readthedocs.io/en/2.1.0/_modules/imapclient/imapclient.html#IMAPClient.xlist_folders

        See Also:
            Uses broad ``Exception`` clause to catch login errors, since the same is raised by ``imaplib``
        """
        if folder not in self.FOLDERS:
            raise ValueError(
                f"The chosen folder {folder} is not available.\n"
                f"Available Options: {', '.join(self.FOLDERS)}"
            )
        self.folder = folder
        self.mail = None
        if not all([gmail_user, gmail_pass]):
            raise ValueError(
                'Cannot proceed without the args or env vars: `gmail_user` and `gmail_pass`'
            )
        self.mail = IMAP4_SSL('imap.gmail.com')  # connects to gmail using imaplib
        # noinspection PyBroadException
        try:
            self.mail.login(user=gmail_user, password=gmail_pass)  # login to your gmail account using the env variables
            self.mail.list()  # list all the folders within your mailbox (like inbox, sent, drafts, etc)
            self.mail.select(folder)
        except Exception:
            self.mail = None
        self.gmail_user = gmail_user

    def instantiate(self, category: str = 'UNSEEN') -> Response:
        """Searches the number of emails for the category received and forms.

        Args:
            category: Search criteria to be performed.

        References:
            - Categories, ``SMALLER``, ``TEXT``, and ``SINCE`` take additional values.

        Examples:
            ``'SMALLER 500'``
            ``'TEXT "foo-bar"'``
            ``'SINCE 17-Feb-2022'`` [OR] ``'SINCE 1645300995.231752'``

        References:
            https://imapclient.readthedocs.io/en/2.1.0/api.html#imapclient.IMAPClient.search

        Returns:
            Response:
            A Response class containing number of email messages, return code and the encoded messages itself.
        """
        if not self.mail:
            return Response(dictionary={
                'ok': False,
                'status': 401,
                'body': 'BUMMER! Unable to read your emails.\n\nTroubleshooting Steps:\n'
                        '\t1. Make sure your username and password are correct.\n'
                        '\t2. Logon to https://myaccount.google.com/lesssecureapps and turn ON less secure apps.\n'
                        '\t3. If you have enabled 2 factor authentication, use thee App Password generated.'
            })

        if category not in self.CATEGORIES and not any(cat in category for cat in self.ADDITIONAL_CATEGORIES):
            raise ValueError(
                f"The chosen search criteria {category} is not available.\n"
                f"Available Options: {', '.join(self.CATEGORIES)}"
            )

        return_code, messages = self.mail.search(None, category)
        if return_code != 'OK':
            return Response(dictionary={
                'ok': False,
                'status': 404,
                'body': 'Unable to read emails.'
            })

        num = len(messages[0].split())
        if not num:
            return Response(dictionary={
                'ok': False,
                'status': 204,
                'body': f'No {category.lower()} emails found for {self.gmail_user} in {self.folder}',
                'count': num
            })

        if return_code == 'OK':
            return Response(dictionary={
                'ok': True,
                'status': 200,
                'body': messages,
                'count': num
            })

    def _get_info(self, response_part: tuple) -> dict:
        """Extracts sender, subject and time received from response part.

        Args:
            response_part: Encoded tuple of the response part in the email.

        Returns:
            dict:
            A dictionary of sender, subject and received time.
        """
        original_email = message_from_bytes(response_part[1])
        if received := original_email.get('Received'):
            date = received.split(';')[-1].strip()
        else:
            date = original_email.get('Date')
        if '(PDT)' in date:
            datetime_obj = datetime.strptime(date, "%a, %d %b %Y %H:%M:%S -0700 (PDT)")
        elif '(PST)' in date:
            datetime_obj = datetime.strptime(date, "%a, %d %b %Y %H:%M:%S -0800 (PST)")
        else:
            datetime_obj = datetime.now()
        sender = make_header(decode_header((original_email['From']).split(' <')[0]))
        sub = make_header(decode_header(original_email['Subject'])) \
            if original_email['Subject'] else None
        # Converts pacific time to local timezone as the default is pacific
        local_time = datetime_obj.replace(tzinfo=pytz.timezone('US/Pacific')).astimezone(tz=self.LOCAL_TIMEZONE)
        received_date = local_time.strftime("%Y-%m-%d")
        current_date_ = datetime.today().date()
        # Replaces current date with today or yesterday to make it simpler
        if received_date == str(current_date_):
            receive = local_time.strftime("Today, at %I:%M %p")
        elif received_date == str(current_date_ - timedelta(days=1)):
            receive = local_time.strftime("Yesterday, at %I:%M %p")
        else:
            receive = local_time.strftime("on %A, %B %d, at %I:%M %p")
        return {'sender': str(sender).strip(), 'subject': str(sub).strip(), 'datetime': receive.strip()}

    def read_email(self, messages: list or str) -> str:
        """Prints unread emails one by one after getting user confirmation.

        Args:
            Takes the encoded message list as an argument. This is the body of the ``instantiate`` method.

        Returns:
            Response:
            A custom response class with properties: ok, status and body to the user.
        """
        for nm in messages[0].split():
            dummy, data = self.mail.fetch(nm, '(RFC822)')
            for each_response in data:
                if isinstance(each_response, tuple):
                    yield self._get_info(response_part=each_response)
        else:  # Executes when there is no break statement in the loop above
            if self.mail:
                self.mail.close()
                self.mail.logout()


if __name__ == '__main__':
    reader = ReadEmail(folder='"[Gmail]/All Mail"')
    response = reader.instantiate(category='SMALLER 500')
    if response.ok:
        unread_emails = reader.read_email(response.body)
        for each_mail in list(unread_emails):
            print(each_mail)
    else:
        print(response.body)



import socket
from typing import Any, List


def hostname_to_ip(hostname: str) -> List[str]:
    """Uses ``socket.gethostbyname_ex`` to translate a host name to IPv4 address format, extended interface.

    References:
        https://docs.python.org/3/library/socket.html#socket.gethostbyname_ex

    Args:
        hostname: Takes the hostname of a device as an argument.
    """
    try:
        _hostname, _alias_list, _ipaddr_list = socket.gethostbyname_ex(hostname)
    except socket.error:
        return []
    return _ipaddr_list


def matrix_to_flat_list(input_: List[list]) -> list:
    """Converts a matrix into flat list.

    Args:
        input_: Takes a list of list as an argument.

    Returns:
        list:
        Flat list.
    """
    return sum(input_, []) or [item for sublist in input_ for item in sublist]


def remove_duplicates(input_: List[Any]) -> List[Any]:
    """Remove duplicate values from a list.

    Args:
        input_: Takes a list as an argument.

    Returns:
        list:
        Returns a cleaned up list.
    """
    # return list(set(input_))
    return [i.strip() for n, i in enumerate(input_) if i not in input_[n + 1:]]



import re
import smtplib
import socket
from typing import Union

from gmailconnector.responder import Response

from .address import ValidateAddress
from .domain import get_mx_records
from .exceptions import AddressFormatError, InvalidDomain, NotMailServer
from .port import is_port_allowed


def validate_email(email_address: str, smtp_timeout: Union[int, float] = 5, sender: str = None) -> Response:
    """Validates email address deliver-ability using SMTP.

    Args:
        email_address: Email address.
        smtp_timeout: Time in seconds to wait for a result.
        sender: Sender's email address.

    Warnings:
        - This is not a perfect solution for validating email address.
        - Works perfect for gmail, but yahoo and related mail servers always returns OK even with garbage email address.
        - Works only if port 25 is not blocked by ISP.
        - SMTP Timeout ``(smtp_timeout)`` specified is to create a socket connection with each mail exchange server.
        - If a mail server has 10 mx records and timeout is set to 3, the total wait time will be 30 seconds.

    See Also:
        - Sets the ``ok`` flag in Response class to
            - ``False`` only if the email address or domain is clearly invalid.
            - ``True`` only if the email address is clearly valid.
            - ``None`` if port 25 is blocked or all mx records returned temporary errors.

    Returns:
        bool:
        Boolean flag to indicate if the email address is valid.
    """
    try:
        address = ValidateAddress(address=email_address)
    except AddressFormatError as error:
        return Response(dictionary={
            'ok': False,
            'status': 422,
            'body': f"Invalid address: {email_address!r}. {error}" if str(error).strip() else
            f"Invalid address: {email_address!r}."
        })

    try:
        mx_records = get_mx_records(domain=address.domain)
    except InvalidDomain as error:
        return Response(dictionary={
            'ok': False,
            'status': 422,
            'body': error.__str__()
        })
    except NotMailServer as error:
        return Response(dictionary={
            'ok': False,
            'status': 422,
            'body': error.__str__()
        })

    if not is_port_allowed():
        return Response(dictionary={
            'ok': None,
            'status': 305,
            'body': 'Cannot verify SMTP since port 25 has been blocked.'
        })

    server = smtplib.SMTP(timeout=smtp_timeout)
    for record in mx_records:
        try:
            server.connect(host=record)
        except socket.error:
            continue
        server.ehlo_or_helo_if_needed()
        server.mail(sender=sender)
        code, msg = server.rcpt(recip=address.email)
        msg = re.sub(r"\d+.\d+.\d+", '', msg.decode(encoding='utf-8')).strip()
        msg = ' '.join(msg.splitlines()).replace('  ', ' ').strip() if msg else "Unknown error"
        if code == 550:  # Definitely invalid email address
            return Response({
                'ok': False,
                'status': 550,
                'body': msg
            })
        if code < 400:  # Valid email address
            return Response({
                'ok': True,
                'status': 200,
                'body': msg
            })
    return Response({
        'ok': None,
        'status': 207,
        'body': 'Received multiple temporary errors. Could not finish validation.'
    })



import os

from setuptools import setup

from version import version_info

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Telecommunications Industry',
    'Operating System :: MacOS :: MacOS X',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.9',
    'Topic :: Communications :: Email :: Post-Office :: IMAP'
]


def read(name):
    """https://pythonhosted.org/an_example_pypi_project/setuptools.html#setting-up-setup-py - reference."""
    return open(os.path.join(os.path.dirname(__file__), name)).read()


setup(
    name='gmail-connector',
    version='.'.join(str(c) for c in version_info),
    description='Python module to, send SMS, emails and read unread emails.',
    long_description=read('README.md') + '\n\n' + read('CHANGELOG'),
    long_description_content_type='text/markdown',
    url='https://github.com/thevickypedia/gmail-connector',
    author='Vignesh Sivanandha Rao',
    author_email='svignesh1793@gmail.com',
    License='MIT',
    classifiers=classifiers,
    keywords='gmail, smtp, imap, tls',
    packages=['.gmailconnector'],
    python_requires=">=3.8",
    install_requires=['idna>=3.3', 'setuptools>=58.2.0', 'python-dotenv>=0.21.0', 'pytz>=2021.3'],
    project_urls={
        'Docs': 'https://thevickypedia.github.io/gmail-connector',
        'Bug Tracker': 'https://github.com/thevickypedia/gmail-connector/issues'
    },
)



import smtplib
import socket


def is_port_allowed(host: str = 'gmail.com', port: int = smtplib.SMTP_PORT, timeout: int = 3):
    """Replicates ``telnet`` command by trying to create a socket connection on a particular host and port.

    Args:
        host: Host that has to be tried. Defaults to 'gmail.com'
        port: Port number that has to be checked. Defaults to SMTP port 25.
        timeout: Timeout for socket connection.

    Returns:
        bool:
        Boolean flag to indicate whether the port is allowed.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    if sock.connect_ex((host, port)) == 0:
        sock.close()
        return True
    sock.close()



import os
from email.mime import multipart, text
from email.mime.application import MIMEApplication
from smtplib import SMTP, SMTPAuthenticationError, SMTPConnectError
from typing import Union

from dotenv import load_dotenv

from .responder import Response

if os.path.isfile('.env'):
    load_dotenv(dotenv_path='.env')


class SendEmail:
    """Initiates Emailer object to email defined recipient from a defined sender with or without attachments.

    >>> SendEmail

    """

    def __init__(self, gmail_user: str = os.environ.get('gmail_user') or os.environ.get('GMAIL_USER'),
                 gmail_pass: str = os.environ.get('gmail_pass') or os.environ.get('GMAIL_PASS')):
        """Initiates necessary args, creates a connection with Gmail's SMTP on port 587.

        Args:
            gmail_user: Login email address.
            gmail_pass: Login password.
        """
        self.server = None
        if not all([gmail_user, gmail_pass]):
            raise ValueError(
                'Cannot proceed without the args: `gmail_user` and `gmail_pass`'
            )
        self.gmail_user = gmail_user
        self.gmail_pass = gmail_pass
        self.server = SMTP(host='smtp.gmail.com', port=587)
        self._authenticated = False
        self._failed_attachments = {"FILE NOT FOUND": [], "FILE SIZE OVER 25 MB": []}

    @property
    def authenticate(self) -> Response:
        """Starts the TLS server and authenticates the user.

        Returns:
            Response:
            A custom response class with properties: ok, status and body to the user.
        """
        self.server.starttls()
        try:
            self.server.login(user=self.gmail_user, password=self.gmail_pass)
            self._authenticated = True
            return Response(dictionary={
                'ok': True,
                'status': 200,
                'body': "Authentication was successful."
            })
        except SMTPAuthenticationError:
            self.server = None
            return Response(dictionary={
                'ok': False,
                'status': 401,
                'body': "GMAIL login failed with SMTPAuthenticationError: Username and Password not accepted.\n"
                        "Ensure the credentials stored in env vars are set correct.\n"
            })
        except SMTPConnectError:
            self.server = None
            return Response(dictionary={
                'ok': False,
                'status': 503,
                'body': "Error during connection establishment with GMAIL server."
            })

    def __del__(self):
        """Destructor has been called to close the TLS connection and logout."""
        if self.server:
            self.server.close()

    def _multipart_message(self, subject: str, recipient: str or list, sender: str, body: str, html_body: str,
                           attachments: list, filenames: list, cc: str or list) -> multipart.MIMEMultipart:
        """Creates a multipart message with subject, body, from and to address, and attachment if filename is passed.

        Args:
            recipient: Email address of the recipient to whom the email has to be sent.
            subject: Subject line of the email.
            body: Body of the email. Defaults to ``None``.
            html_body: Body of the email. Defaults to ``None``.
            attachments: Names of the files that has to be attached.
            filenames: Custom names of the attachments.
            cc: Email address of the recipient to whom the email has to be CC'd.
            sender: Add sender name to the email.

        Returns a message if a filename is given for attachment but not available at the given path.

        Returns:
            `multipart.MIMEMultipart`:
            MIMEMultipart version of the created message.
        """
        recipient = [recipient] if isinstance(recipient, str) else recipient
        cc = [cc] if cc and isinstance(cc, str) else cc

        msg = multipart.MIMEMultipart()
        msg['Subject'] = subject
        msg['From'] = f"{sender} <{self.gmail_user}>"
        msg['To'] = ','.join(recipient)
        if cc:
            msg['Cc'] = ','.join(cc)

        if body:
            msg.attach(payload=text.MIMEText(body, 'plain'))
        if html_body:
            msg.attach(payload=text.MIMEText(html_body, 'html'))

        for index, attachment_ in enumerate(attachments):
            file_type = attachment_.split('.')[-1]
            try:
                filename = filenames[index]
            except IndexError:
                filename = None
            if filename and '.' in filename:  # filename is passed with an extn
                pass
            elif filename and '.' in attachment_:  # file name's extn is got from attachment name
                filename = f'{filename}.{file_type}'
            elif filename:  # filename is passed without an extn so proceeding with the same
                pass
            else:
                filename = attachment_.split(os.path.sep)[-1].strip()  # rips path from attachment as filename

            if not os.path.isfile(attachment_):
                self._failed_attachments["FILE NOT FOUND"].append(filename)
                continue
            if os.path.getsize(attachment_) / 1e+6 > 25:
                self._failed_attachments["FILE SIZE OVER 25 MB"].append(filename)
                continue

            with open(attachment_, 'rb') as file:
                attribute = MIMEApplication(file.read(), _subtype=file_type)
            attribute.add_header('Content-Disposition', 'attachment', filename=filename)
            msg.attach(payload=attribute)

        return msg

    def send_email(self, subject: str,
                   recipient: Union[str, list] = os.environ.get('recipient') or os.environ.get('RECIPIENT'),
                   sender: str = 'GmailConnector', body: str = None, html_body: str = None,
                   attachment: Union[str, list] = None, filename: Union[str, list] = None,
                   custom_attachment: dict[str, str] = None, cc: Union[str, list] = None, bcc: Union[str, list] = None,
                   fail_if_attach_fails: bool = True) -> Response:
        """Initiates a TLS connection and sends the email.

        Args:
            recipient: Email address of the recipient to whom the email has to be sent.
            subject: Subject line of the email.
            body: Body of the email. Defaults to ``None``.
            html_body: Body of the email. Defaults to ``None``.
            attachment: Name of the file that has to be attached.
            filename: Custom name of the attachment.
            custom_attachment: Dictionary of custom name for the attachment as key and the relative filepath as value.
            cc: Email address of the recipient to whom the email has to be CC'd.
            bcc: Email address of the recipient to whom the email has to be BCC'd.
            sender: Add sender name to the email.
            fail_if_attach_fails: Boolean flag to restrict sending the email if attachment is included but fails.

        Returns:
            Response:
            A custom response class with properties: ok, status and body to the user.
        """
        if not self._authenticated:
            status = self.authenticate
            if not status.ok:
                return status

        if custom_attachment:
            attachments = list(custom_attachment.keys())
            filenames = list(custom_attachment.values())
        else:
            attachments = [attachment] if isinstance(attachment, str) else attachment if attachment else []
            filenames = [filename] if isinstance(filename, str) else filename if filename else []

        msg = self._multipart_message(subject=subject, sender=sender, recipient=recipient, attachments=attachments,
                                      body=body, html_body=html_body, cc=cc, filenames=filenames)

        unattached = {k: ', '.join(v) for k, v in self._failed_attachments.items() if v}
        if fail_if_attach_fails and unattached:
            return Response(dictionary={
                'ok': False,
                'status': 422,
                'body': f"Email was not sent. Unattached: {unattached!r}"
            })

        recipients = [recipient] if isinstance(recipient, str) else recipient
        if cc:
            recipients.append(cc) if isinstance(cc, str) else recipients.extend(cc)
        if bcc:
            recipients.append(bcc) if isinstance(bcc, str) else recipients.extend(bcc)
        self.server.sendmail(
            from_addr=sender,
            to_addrs=recipients,
            msg=msg.as_string()
        )
        if unattached:
            return Response(dictionary={
                'ok': True,
                'status': 206,
                'body': f"Email has been sent to {recipient!r}. Unattached: {unattached!r}."
            })
        else:
            return Response(dictionary={
                'ok': True,
                'status': 200,
                'body': f"Email has been sent to {recipient!r}"
            })



from concurrent.futures import ThreadPoolExecutor
from email import message_from_bytes
from email.header import decode_header, make_header
from imaplib import IMAP4_SSL


class DeleteSent:
    """Initiates DeleteSent object to delete a particular email from SentItems.

    >>> DeleteSent

    """

    def __init__(self, username: str, password: str, subject: str):
        self.mail = IMAP4_SSL('imap.gmail.com')
        self.username = username
        self.subject = subject
        self.mail.login(user=username, password=password)

    def _thread_executor(self, item_id: bytes or str) -> None:
        """Gets invoked in multiple threads, to set the flag as ``Deleted`` for the message which was just sent.

        Args:
            item_id: Takes the ID of the message as an argument.
        """
        dummy, data = self.mail.fetch(item_id, '(RFC822)')
        for response_part in data:
            if not isinstance(response_part, tuple):
                continue
            original_email = message_from_bytes(response_part[1])  # gets the raw content
            sender = str(make_header(decode_header((original_email['From']).split(' <')[0])))
            sub = str(make_header(decode_header(original_email['Subject'])))
            to = str(make_header(decode_header(original_email['To'])))
            if to == to and sub == self.subject and sender == self.username:
                self.mail.store(item_id.decode('UTF-8'), '+FLAGS', '\\Deleted')
                self.mail.expunge()
                self.mail.close()
                self.mail.logout()
                break

    def delete_sent(self):
        """Deletes the email from GMAIL's sent items right after sending the message.

        See Also:
            Invokes ``thread_executor`` with 20 workers to sweep through sent items to delete the email.
        """
        self.mail.list()
        self.mail.select('"[Gmail]/Sent Mail"')
        return_code, messages = self.mail.search(None, 'ALL')  # Includes SEEN and UNSEEN, although sent is always SEEN
        if return_code != 'OK':
            return
        with ThreadPoolExecutor(max_workers=1) as executor:
            executor.map(self._thread_executor, sorted(messages[0].split(), reverse=True))
